;TODOs:
; + prefer turn off
; + correct x-move
; + optimize CheckRight/-Left bzw. -Up/-Down  (merge!)
; + reverse X/Y in CheckLeft/Right etc.
; o Tank AI:
;   + find way through labyrinth
;   o evade other tanks (TODO: thinner area)
;   + attack player (TODO: improve randomness)
;   x revive AI (>1 tank disabled)
; x player shot aligned to grid
; + PF gets priority
; + better missiles
; + tank not killed by own missile
; o less strict tank shield direction check (TODO: try to optimize)
; o kill helicopter (TODO: stop counters)
; x copy graphics instead of PF into RAM
; - sounds
; + revive stunned tanks (by other tank) (bugfixed)
; - mega-boss
; - helicopter damages (speed, gun, radar, fuel leak)
; - gun recoil
; o fuel (TODO: consumption)
; - bonus items (extra or instead of tanks)
; ? multi hits at tank shield
; ? inertia (Paradroid like)
; ? scroll to next maze
; + generate random maze (DONE: better random generator)
; - correct SELECT and RESET
; ? vertically adjust missiles (VDELPx?)
; - ...


    processor 6502
    include vcs.h

;===============================================================================
; A S S E M B L E R - S W I T C H E S
;===============================================================================

DISABLE     = 0             ; disable code (do NOT change)
DEBUG       = 1             ; debug mode
STACKCHECK  = 1             ; check stack space
ILLEGAL     = 0             ; use illegal opcodes
PREDEFINED  = 0


;===============================================================================
; T I A - C O N S T A N T S
;===============================================================================

;ENAxy:
ENABLE      = %10

;REFPx:
GR_REFLECT  = %1000

;CTRLPF:
PF_REFLECT  = %001
PF_SCORE    = %010
PF_PRIORITY = %100

;NUSIZx:
MSBL_SIZE1  = %000000
MSBL_SIZE2  = %010000
MSBL_SIZE4  = %100000
MSBL_SIZE8  = %110000


;===============================================================================
; C P U - C O N S T A N T S
;===============================================================================

opBIT_B     = $24
opBIT_W     = $2c

opNOP_B     = $82   ; 2 cycles
opNOP_W     = $0c
opNOP_ZP    = $04   ; 3 cycles

;===============================================================================
; C O N S T A N T S
;===============================================================================

NUM_MAZES   = 8
NUM_XBLOCKS = 9
NUM_YBLOCKS = 7
NUM_TANKS   = 4

H_VERT      = 20/2
H_HORZ      =  6/2
NUM_BLOCKS  = H_VERT+H_HORZ         ; = 13

MAXDIR      = 16
DIR_UP      = %0000
DIR_RIGHT   = %0100
DIR_DOWN    = %1000
DIR_LEFT    = %1100
DIR_HORZ    = DIR_RIGHT
DIR_HIGH    = DIR_DOWN

ENABLE_TURN = $80
NEG_TURN    = $40|ENABLE_TURN
POS_TURN    = $00|ENABLE_TURN
NO_TANK     = $20
U_TURN      = $10
TURN_FLAGS  = NEG_TURN|U_TURN

H_PLAYER    =  7
H_TANK      =  7
H_SHOT      =  2
H_SCORE     =  5

ID_P0       =  0
ID_P1       =  1
ID_HELI     =  2
ID_MISL     =  3
ID_SHOT     =  4
ID_TANK2    =  5
ID_TANK3    =  6
ID_TANK0    =  7
ID_TANK1    =  8
ID_TANKLST  =  5

X0          = 15
Y0          = 10

Y_TANK0     = Y0    ; green
Y_TANK1     = Y0    ; orange
Y_TANK2     = Y0    ; blue
Y_TANK3     = Y0    ; red       ;Y0+13*4;

X_TANK0     = X0+24+2
X_TANK1     = X0
X_TANK2     = X0+48
X_TANK3     = X0+69+5           ;X0+92+1 ;

DIR_TANK0   = DIR_LEFT
DIR_TANK1   = DIR_LEFT
DIR_TANK2   = DIR_RIGHT
DIR_TANK3   = DIR_RIGHT         ;(NO_TANK|U_TURN|POS_TURN)|%101    ;

X_PLAYER    = X0+124
Y_PLAYER    = Y0+78
X_GRID      = NUM_XBLOCKS*2-2
Y_GRID      = NUM_YBLOCKS*2-2

NUM_SPEEDS  = 4

FUEL_COL    = $ca
LIVES_COL   = $ae                   ; ~ helicopter color

MAX_FUEL    = 115

RAND_EOR_8   = $b2
RAND_EOR_16  = $b9a2                 ; asl %1011100110100010
;RAND_EOR    = $b39d                 ; lsr


;===============================================================================
; Z P - V A R I A B L E S
;===============================================================================

    SEG.U   variables
    ORG     $80

random      .byte                   ; random number
randomLo    .byte
randomHi    .byte
frameCnt    .byte                   ; frame counter
maze        .byte                   ; number of current maze
;level       .byte                   ; difficulty level
score       .word                   ; BCD score
lives       .byte                   ; remaining lives pattern (0x0x0x0x)
fuel        .byte                   ; remaining fuel 0..115

PF0Lst      ds NUM_BLOCKS           ; kernel
PF1Lst      ds NUM_BLOCKS           ; kernel
PF2Lst      ds NUM_BLOCKS           ; kernel        next byte is overwritten in DecodePF !!!

;---------------------------------------
ptrP0       .word                   ; kernel temp
ptrP1       .word                   ; kernel temp
ptrBLSize   .word                   ; kernel temp
ptrBLMove   .word                   ; kernel temp

hMisl       .byte                   ; kernel

tmpVar      .word                   ; temp
tmpVar2     = tmpVar+1              ; temp
tmpVar3     = ptrP0                 ; temp
tmpVar4     = tmpVar3+1             ; temp
jmpPtr      = ptrP1                 ; temp
tmpVar5     = ptrBLSize             ; temp
tmpVar6     = tmpVar5+1             ; temp
tmpVar7     = ptrBLMove             ; temp
tmpVar8     = tmpVar7+1             ; temp

tmp0        .byte
tmp1        .byte
tmp2        .byte
tmp3        .byte
;---------------------------------------
xPosLst     ds 5+4
xPosP0      = xPosLst+0
xPosP1      = xPosLst+1
xPosPlayer  = xPosLst+2
xPosMisl    = xPosLst+3
xPosShot    = xPosLst+4

xPosTankLst = xPosLst+5
xPosTank0   = xPosTankLst+0
xPosTank1   = xPosTankLst+1
xPosTank2   = xPosTankLst+2
xPosTank3   = xPosTankLst+3
;---------------------------------------
yPosLst     ds 5+4
yPosP0      = yPosLst+0             ; kernel
yPosP1      = yPosLst+1             ; kernel
yPosPlayer  = yPosLst+2
yPosMisl    = yPosLst+3             ; kernel
yPosShot    = yPosLst+4             ; kernel

yPosTankLst = yPosLst+5
yPosTank0   = yPosTankLst+0
yPosTank1   = yPosTankLst+1
yPosTank2   = yPosTankLst+2
yPosTank3   = yPosTankLst+3

yPosHeli    .byte                   ; kernel temp   (copied from yPosPlayer)

;---------------------------------------
dirPlayer   ds 3+4
dirLst      = dirPlayer-2
dirMisl     = dirLst+3
dirShot     = dirLst+4

dirTankLst  = dirLst+5

mislTank    .byte                   ; id of missile fireing tank

;---------------------------------------
xGrid       .byte
yGrid       .byte       ;
xyMove      .byte
xyDir       .byte       ; y-000vvv

killCnt     .byte


speedLst0       ds 4
speedLst        = speedLst0-2
speedPlayer     = speedLst+2
speedMisl       = speedLst+3
speedShot       = speedLst+4
speedTank       = speedLst+5

speedLstHi0     ds 4
speedLstHi      = speedLstHi0-2
speedPlayerHi   = speedLstHi+2          ;   always 0!
speedMislHi     = speedLstHi+3
speedShotHi     = speedLstHi+4
speedTankHi     = speedLstHi+5          ;   always 0!

delayLst0       ds 4
delayLst        = delayLst0-2
delayPlayer     = delayLst+2
delayMisl       = delayLst+3
delayShot       = delayLst+4
delayTank       = delayLst+5

; PF pattern for fuelbar
fuelPF0l        = tmpVar4
fuelPF1l        .byte
fuelPF2l        .byte
fuelPF0r        .byte
fuelPF1r        .byte
hMoveBL         .byte
saveCXBLPF      = hMoveBL
fuelCol         = tmpVar2
levelCol        = tmpVar3


  IF STACKCHECK
stackError  .byte
  ENDIF


;===============================================================================
; M A C R O S
;===============================================================================

  MAC BIT_B
    .byte   opBIT_B
  ENDM

  MAC BIT_W
    .byte   opBIT_W
  ENDM

  MAC NOP_B
    IF ILLEGAL
      .byte   opNOP_B
    ELSE
      BIT_B
    ENDIF
  ENDM

  MAC NOP_W
    IF ILLEGAL
      .byte   opNOP_W
    ELSE
      BIT_W
    ENDIF
  ENDM

  MAC NOP_ZP
    IF ILLEGAL
      .byte   opNOP_ZP
    ELSE
      BIT_B
    ENDIF
  ENDM


  MAC DEBUG_BRK
    IF DEBUG
      brk                         ;
    ENDIF
  ENDM

  MAC SLEEP
    IF {1} = 1
      ECHO "ERROR: SLEEP 1 not allowed !"
      END
    ENDIF
    IF {1} & 1
      NOP_ZP
      .byte $00
      REPEAT ({1}-3)/2
        nop
      REPEND
    ELSE
      REPEAT ({1})/2
        nop
      REPEND
    ENDIF
  ENDM


;===============================================================================
; R O M - C O D E
;===============================================================================

    SEG     Bank0
    ORG     $f000, $ff

Kernel SUBROUTINE
    ds 1,0

DrawScreen:
.waitTim:
    ldx     INTIM
    bne     .waitTim
    sta     WSYNC
;---------------------------------------------------------------
    stx     VBLANK

; *** draw top wall: ***
    dex
    stx     PF1
    stx     PF2
    lda     #$c3                ;           lover bits used for ENAM0
    sta     PF0

    ldx     #H_HORZ*2-1
.loopTop:
    sta     WSYNC
;---------------------------------------------------------------
    dex                         ; 2
    bne     .loopTop            ; 2

; *** prepare entering the main kernel: ***
    ldy     #7*H_VERT + 6*H_HORZ+1;2
    ldx     #NUM_BLOCKS-1       ; 2
    nop                         ; 2
    BIT_B                       ; 1
.skipDrawPlayer:                ;           @11
    nop                         ; 2
    lsr                         ; 2
    sta     ENAM0               ; 3
    bcs     .contDrawPlayer     ; 3 = 10

; ***   K E R N E L   ***
.loopBlock:                     ;           @51
    nop                         ; 2
    nop                         ; 2
.nextBlock:                     ;           @55
    lda     PF2Lst,x            ; 4
    sta     PF2                 ; 3         @62     >=59
    lda     PF1Lst,x            ; 4
    sta     PF1                 ; 3         @69     >=69
    lda     PF0Lst,x            ; 4
.cont:
    sta     PF0                 ; 3 = 21    @76     >=76
;---------------------------------------

    tya                         ; 2                 draw tank missile
    sbc     yPosMisl            ; 3
    adc     hMisl               ; 3
    lda     #ENABLE             ; 2
    sbc     #0                  ; 2
    sta     ENAM1               ; 3 = 15    @15

    tya                         ; 2                 draw player shot
    sbc     yPosShot            ; 3
    adc     #H_SHOT             ; 2
    lda     #ENABLE             ; 2
    sbc     #0                  ; 2
    sta     ENABL               ; 3 = 14    @29     (VDELed!)

    tya                         ; 2                 draw 2nd tank (part 1/2)
    sbc     yPosP1              ; 3
    adc     #H_TANK             ; 2
    lda     #0                  ; 2
    bcc     .skipDrawP1         ; 2
    lda     (ptrP1),y           ; 5
.skipDrawP1:
    sec                         ; 2
    pha                         ; 3 = 17/21 @50

    tya                         ; 2                 draw helicopter (part 1/2)
    sbc     yPosHeli            ; 3
    adc     #H_PLAYER           ; 2
    bcc     .skipDrawPlayerA    ; 2
    lda     (ptrBLMove),y       ; 5
    sta     HMM0                ; 3         @67
.skipDrawPlayerA:               ;   = 10/17

    pla                         ; 4                 draw 2nd tank (part 2/2)
    sta     WSYNC               ; 3 =  7    @74
;---------------------------------------
    sta     HMOVE               ; 3
    sta     GRP1                ; 3 =  6    @06

    lda     #ENABLE|1           ; 2                 draw helicopter (part 2/2)
    bcc     .skipDrawPlayer     ; 2
    sta     ENAM0               ; 3 =  7    @13
    lda     (ptrBLSize),y       ; 5
    sta     NUSIZ0              ; 3 =  8    @21
.contDrawPlayer:                ;

    dey                         ; 2
    tya                         ; 2                 draw 1st tank
    sbc     yPosP0              ; 3
    adc     #H_TANK             ; 2
    bcs     .doDrawP0           ; 2
    lda     #0                  ; 2
    NOP_W                       ;-1
.doDrawP0:
    lda     (ptrP0),y           ; 5
    sta.w   GRP0                ; 4 = 21/21 @42     (VDELed!)

    tya                         ; 2
    cmp     BlockSizeTbl,x      ; 4                 current PF block finished?
    bne     .loopBlock          ; 2=  8/ 9
    dex                         ; 2
    bpl     .nextBlock          ; 2=  4/ 5 @55

;**********  F U E L   B A R  **********
    lda     #LIVES_COL          ; 2
    sta     COLUP0              ; 3
    lda     hMoveBL             ; 3
    ldy     CXBLPF              ; 3
    sty     saveCXBLPF          ; 3                 = hMoveBL!
    sta     WSYNC               ; 3 = 17    @72
;---------------------------------------------------------------
    ldy     #$c0                ; 2
    sty     PF0                 ; 3
    stx     PF1                 ; 3
    stx     PF2                 ; 3
    sta     HMBL                ; 3
    and     #$0f                ; 2
    tay                         ; 2 = 18
.waitBL:                        ;
    dey                         ; 2
    bpl     .waitBL             ; 2
    lda     #$e0                ; 2
    sta     RESBL               ; 3 =  9    @27..62
    sta     HMP0                ; 3
                                ; 8 free cycles here

    sta     WSYNC               ; 3
;---------------------------------------------------------------
    lda     #$a0                ; 2
    sta.w   HMM1                ; 4
    lda     #%000101            ; 2
    sta     NUSIZ0              ; 3
    lda     #%100111            ; 2
    sta     NUSIZ1              ; 3
    sta     REFP0               ; 3
    sta     RESM1               ; 3         @22
    sta     REFP1               ; 3

    lda     randomHi            ; 3
    and     #$f7                ; 2
    adc     #2                  ; 2                 don't care for CF
    sta     levelCol            ; 3
    sta     COLUP1              ; 3

    lda     fuel                ; 3
    cmp     #4                  ; 2
    and     #3                  ; 2
    tay                         ; 2
    ror                         ; 2
    sta     fuelPF0l            ; 3

    lda     #$90                ; 2
    sta.w   HMP1                ; 4

    sta     RESP1               ; 3         @61
    sta     RESP0               ; 3         @64

    ldx     lives               ; 3
    lda     FuelBLSizeTbl,y     ; 4
    sta     CTRLPF              ; 3
    lda     #$fc >> 1           ; 2                 missile must be enabled that late,
;---------------------------------------------------------------
    sta     HMOVE               ; 3         @03
    sta     ENAM1               ; 3                  else hardware collisions won't work
    ldy     #0                  ; 2
    sty     PF0                 ; 3
    sty     PF1                 ; 3
    sty     PF2                 ; 3         @17

    sty     VDELBL              ; 3                 disable VDELed BL
    stx     GRP0                ; 3
    asl                         ; 2                 a = $fc!
    sta     GRP1                ; 3         @28

    ldx     #1                  ; 2
    sta     HMCLR               ; 3         @33

; --------------------- PF Timings --------------------
; | PF0 |   PF1   |   PF2   | PF0 |   PF1   |   PF2   |
; |22.??|27 ..  ??|38 ..  ??|48.51|54 ..  ??|64 ..  ??|

.loopFuel:
    sta     WSYNC               ; 3
;---------------------------------------------------------------
    sta     HMOVE               ; 3         @03
    lda     #2                  ; 2                 missile must be enabled that late,
    sta     ENABL               ; 3                  else hardware collisions won't work
    lda     fuelCol             ; 3
    sta     COLUPF              ; 3         @14
    lda     fuelPF0l            ; 3
    sta     PF0                 ; 3         @20
    lda     fuelPF1l            ; 3
    sta     PF1                 ; 3         @26
    lda     fuelPF2l            ; 3
    sta     PF2                 ; 3         @32
    lda     fuelPF0r            ; 3
    sta     PF0                 ; 3         @38
    lda     fuelPF1r            ; 3
    sta     PF1                 ; 3         @44
    lda     #$00                ; 2
    sta     PF2                 ; 3         @49
    ldy     #$10                ; 2
    sty     HMBL                ; 3         @54
    dex                         ; 2
    bpl     .loopFuel           ; 2        @58/59

.pageCheck:
  IF >.loopFuel != >.pageCheck
    ECHO ""
    ECHO "ERROR: >.loopFuel != >pageCheck !"
    ECHO ""
    ERR
  ENDIF

    nop
    sta     RESM0               ; 3         @61
    sta     ENABL               ; 3         @64     disable BL
    sta     WSYNC
;---------------------------------------------------------------
    sta     PF0                 ; 3         @03
    sta     PF1                 ; 3
    nop                         ; 2
    ldy     #$0e                ; 2
    sty     ENAM0               ; 3
    sty     COLUP0              ; 3         @16
    ldy     #%001               ;
    sty     NUSIZ0
    sta     RESM0
    nop
    sta     RESM0
    nop
    sta     RESM0
    nop
    sta     RESM0
    nop
    sta     RESM0
    nop
    sta     RESM0
    ldy     #%000101            ; 2
    sta     RESM0
    lda     #LIVES_COL          ; 2
    sta     RESM0
    sty     NUSIZ0              ; 3
    sta     COLUP0              ; 3
    lda     #0
    sta     WSYNC
;--------------------------------------------------------
    sta     ENAM0
    sta     ENAM1               ;                   disable
    ldy     #$c0
    sty     PF0
    stx     PF1                 ; 2                 X = $ff
    stx     PF2

    ldy     levelCol            ; 3
    sty     COLUPF              ; 3
    ldy     #PF_REFLECT         ; 2
    sty     CTRLPF              ; 3

    sta     NUSIZ0              ; 3
    sta     NUSIZ1              ; 3
    sta     GRP0
    sta     GRP1
    sta     WSYNC
    sta     WSYNC
;---------------------------------------
    sta     PF0
    sta     PF1
    sta     PF2
;**********  F U E L   B A R (end)  **********

;*** and now, the (very preliminary) score: ***
    ldy     #PF_SCORE
    sty     CTRLPF

    lda     score
    and     #$0f
    tax
    lda     DigitPtrTbl,x
    sta     tmp0
    lda     score
    lsr
    lsr
    lsr
    lsr
    tax
    lda     DigitPtrTbl,x
    sta     tmp1

    lda     maze
    and     #$07
    tax
    lda     DigitPtrTbl,x
    sta     tmp3


    lda     #$48
    sta     COLUP0
    lda     #$8a
    sta     COLUP1
    ldx     #H_SCORE
.loopScore:
    sta     WSYNC
    ldy     tmp3
    dec     tmp3
    lda     DigitTbl,y
    and     #$0f
    sta     PF1
;    lda     #0
;    sta     PF2

;    lda     #0
;    sta     PF1
    ldy     tmp0
    dec     tmp0
    lda     DigitTbl,y
    and     #$0f
    sta     tmp2

    ldy     tmp1
    dec     tmp1
    lda     DigitTbl,y
    and     #$f0
    ora     tmp2
    sta     PF1

    dex
    bne     .loopScore

    lda     #2
    sta     WSYNC
;---------------------------------------------------------------
    sta     VBLANK
    rts
;Kernel END



Break SUBROUTINE
;    ldx     #frameCnt
;    bne     .contBreak

Start:
    sei                         ;           Disable interrupts, if there are any.
    cld                         ;           Clear BCD math bit.
    ldx     #0
    txs
    pha                         ;           Set stack to beginning.
    txa
.clearLoop:
    pha
    inx
    bne     .clearLoop

    jsr     GameInit

.mainLoop:
    jsr     VerticalBlank
;    jsr     GameCalc
  IF STACKCHECK
    lda     #$a5
    eor     stackError
    beq     .noStackError
    sta     COLUBK
.noStackError:
  ENDIF

    jsr     DrawScreen
    jsr     OverScan
    beq     .mainLoop
    DEBUG_BRK


IncreaseSpeeds SUBROUTINE
    ldx     #NUM_SPEEDS
.loopInc:
    lda     speedLst0,x
    lsr
    lsr
    lsr
    lsr
;    lsr
    adc     speedLst0,x
    sta     speedLst0,x
    lda     speedLstHi0,x
    adc     #0
    sta     speedLstHi0,x
    dex
    bpl     .loopInc
    lda     speedTank
    bpl     .exit
    lda     #$7f
    sta     speedTank
;    inc     speedTankHi
.exit:
    rts


GameInit SUBROUTINE
  IF STACKCHECK
    lda     #$a5
    sta     stackError
  ENDIF
    inc     random
    inc     randomHi
;    lda     #$24
;    sta     randomLo
;    lda     #$b5
;    sta     randomHi

    lda     #$80-1
    sta     speedPlayer
    lda     #$80-1
    sta     speedShot
    lda     #$40-1
    sta     speedTank
    lda     #$80-1
    sta     speedMisl

    ldx     #0
    stx     speedTankHi
    stx     speedPlayerHi
    inx
    stx     speedMislHi
    inx
    stx     speedShotHi

  IF DEBUG
    lda     #0
    sta     maze
;    lda     #$74
;    sta     score
;    lda     #$43
;    sta     score+1
  ENDIF
    lda     #%01010101
    sta     lives

InitMaze SUBROUTINE
  IF PREDEFINED
    jsr     DecodePF
  ELSE
    jsr     CreateMaze
  ENDIF

; reuse x to enable vertical delay:
    stx     VDELP0              ;           X = -1!
;    stx     VDELBL
; TODO: re-enable all tanks

    sec                         ;           set flag for new maze
    NOP_B

InitLife SUBROUTINE
    clc                         ;           set flag for new life
    lda     #X_PLAYER
    sta     xPosPlayer
    lda     #Y_PLAYER
    sta     yPosPlayer
    lda     #X_GRID
    sta     xGrid
    lda     #Y_GRID
    sta     yGrid
    lda     #MAX_FUEL
    sta     fuel

    ldx     #NUM_TANKS-1
.loopPos:
    bcs     .skipCheck          ;           reset all killed tanks in new maze
    lda     dirTankLst,x
    and     #NO_TANK            ;           tank killed?
    bne     .skipRepos          ;            yes, do not reset tank
.skipCheck:
    lda     TankXPos,x
    sta     xPosTankLst,x
    lda     TankYPos,x
    sta     yPosTankLst,x
    lda     TankDir,x
    sta     dirTankLst,x
.skipRepos:
    dex
    bpl     .loopPos
;    rts                         ;          save one byte

GetFrameXY SUBROUTINE
; alternate between displaying tanks 0+1 and 2+3
    lda     #1                  ; 2
    tax                         ; 2
    tay                         ; 2
    bit     frameCnt            ; 3
    beq     .set01              ; 2
    ldy     #3                  ; 2
.set01:
    rts                         ; 6 = 18/19
;GetFrameXY END


CheckMoveObject SUBROUTINE
.speedHi = tmpVar2

    lda     yPosLst,x
    beq     .skipObject

    lda     delayLst,x
    sec
    adc     speedLst,x
    sta     delayLst,x
;    bcc     .skipObject

    lda     speedLstHi,x

;MoveObject:
    ldy     dirLst,x
MoveObjectDir:
; total: 29..43
    sta     .speedHi            ; 3
    lda     XDirTbl,y           ; 4
    beq     .skipX              ; 2
    bpl     .posX               ; 2
    lda     .speedHi            ; 3
    sbc     #0                  ; 2
    eor     #$ff                ; 2
    clc                         ; 2
    NOP_W                       ; 1
.posX:
    lda     .speedHi            ; 3
    adc     xPosLst,x           ; 4
    sta     xPosLst,x           ; 4
    rts                         ; 6 = 29/38

.skipX:                         ;10
    lda     YDirTbl,y           ; 4
    bpl     .posY               ; 2
    lda     .speedHi            ; 3
    sbc     #0                  ; 2
    eor     #$ff                ; 2
    clc                         ; 2
    NOP_W                       ; 1
.posY:
    lda     .speedHi            ; 3
    adc     yPosLst,x           ; 4
    sta     yPosLst,x           ; 4 = 15/28/37
.skipObject:
    rts                         ; 6 =  6
;CheckMoveObject END


VerticalBlank SUBROUTINE
    lda     #3
    sta     WSYNC
    sta     VSYNC

    sta     VDELBL

;*******************************************
    lda     SWCHB
    lsr
    bcs     .skipReset
    brk
.skipReset:
    lsr
    bcs     .skipSelect
    lda     frameCnt
    and     #$1f
    bne     .skipSelect
    ldx     maze
    inx
    cpx     #NUM_MAZES
    bcc     .okMazes
    ldx     #0
.okMazes:
    stx     maze
    jsr     InitMaze
.skipSelect:
    sta     WSYNC
    lda     frameCnt
    bne     .skipDiff
    lda     speedTank
    lsr
    lsr
    lsr
    adc     speedTank
    bcc     .skipMax
    lda     #$ff
.skipMax:
;    sta     speedTank
.skipDiff:
    sta     WSYNC
    lda     #0
    ldx     #44-8
    sta     WSYNC
    sta     VSYNC
    stx     TIM64T
;    rts                         ;               save rts, jsr GameCalc bytes

GameCalc SUBROUTINE
;    lda     frameCnt
;    bne     .skipInc
;    jsr     IncreaseSpeeds
;.skipInc:

; move player shot and tank missile:
    ldx     #ID_SHOT
    jsr     CheckMoveObject
;    jsr     CheckMoveObject
    dex                         ;               x = #ID_MISL
    jsr     CheckMoveObject

; copy values of currently displayed tanks:
    jsr     GetFrameXY          ;               x = 1, y = 1/3
.loopCopyTank:
    lda     xPosTankLst,y
    sta     xPosP0,x
    lda     yPosTankLst,y
    sta     yPosP0,x
    lda     dirTankLst,y
    and     #~TURN_FLAGS        ;               clear turning flags
    sta     tmpVar,x
    lda     TankColTbl,y
    sta     COLUP0,x
    dey
    dex
    bpl     .loopCopyTank

; set low pattern pointers for both displayed tanks:
    ldx     #1
.loopSet:
; let disabled tank blink:
    ldy     #MAXDIR/2+1
    lda     #NO_TANK-1
    cmp     tmpVar,x            ;               tank disabled?
    bcs     .skipDisabled       ;                no, skip
    lda     frameCnt            ;                yes, let disabled tanks blink
    and     #$08
    bne     .contDisabled
.skipDisabled:

; set reflection:
    lda     tmpVar,x            ;               tank direction
    and     #~NO_TANK           ;               clear disabled flag
    cmp     #MAXDIR/2+1
    bcc     .posDir
    eor     #$ff                ;               negate direction
    adc     #MAXDIR             ;               CF=1! (before and after!)
    dec     xPosP0,x            ;               adjust x position
.posDir:
    tay
.contDisabled:
    lda     #GR_REFLECT-1
    adc     #0
    sta     REFP0,x

; select animation pattern:
    lda     frameCnt
    and     #2
    beq     .PatternA
    lda     #H_TANK
.PatternA:
    adc     #<TankPatTbl+H_TANK ;               TODO: optimize
    sbc     yPosP0,x
    adc     Mul14Tbl,y
    sta     ptrP0+1,x
    dex
    bpl     .loopSet
    sta     ptrP0               ;               the last pointer was set wrong!

; set high pattern pointers for tanks and helicopter:
    ldx     #>Tank0a
    stx     ptrP0+1
    stx     ptrP1+1
    dex
    stx     ptrBLSize+1
    stx     ptrBLMove+1

; x-position all five objects:
    ldx     #RESBL-RESP0
.loopXPos:
    lda     xPosLst,x
    jsr     CalcPosX
    jsr     SetPosX
    dex
    bpl     .loopXPos
    sta     WSYNC
    sta     HMOVE

; set low pattern pointers for helicopter:
    lda     dirPlayer
    lsr
    lda     frameCnt
    and     #%10
    beq     .skipB
    lda     #PlayerMoveLeftB-PlayerMoveLeftA
.skipB
    bcc     .dirLeft
    adc     #<PlayerMoveRightA-PlayerMoveLeftA-1
.dirLeft:
    adc     #<PlayerMoveLeftA+H_PLAYER+1
    sbc     yPosPlayer
    sta     ptrBLMove

    lda     #<PlayerSizeTbl+H_PLAYER
;    sec
    sbc     yPosPlayer
    sta     ptrBLSize

; let killed helicopter blink:
    ldx     yPosPlayer
    lda     killCnt
    beq     .skipKill
    lda     frameCnt
    and     #%1000
    bne     .skipKill
    dec     killCnt
    bne     .contKill
    jsr     InitLife
.contKill:
    ldx     #0                  ;               disable helicopter
.skipKill:
    stx     yPosHeli

; set BL size and enable reflected PF:
    lda     #MSBL_SIZE2|PF_PRIORITY|PF_REFLECT
    sta     CTRLPF

; clear all move reigsters:
    sta     HMCLR

; *** calc PF graphics for fuelbar: ***
; clear and set all fuel patterns (except fuelPF0l):
    lda     fuel                ; 3
    lsr                         ; 2
    lsr                         ; 2
    ldy     #0                  ; 2
    ldx     #4                  ; 2 = 11
.loopFuelVal:
    dex                         ; 2
    sty     fuelPF1l,x          ; 4
    cmp     FuelPFValTbl,x      ; 4
    bcc     .loopFuelVal        ; 2= 12/13

    adc     FuelPFOfsTbl,x      ; 4             CF=1!
    tay                         ; 2
    lda     FuelPF1Tbl,y        ; 4 = 10
.loopFuelPF:
    sta     fuelPF1l,x          ; 4
    lda     #$ff                ; 2
    dex                         ; 2
    bpl     .loopFuelPF         ; 2= 10/11
; 76..82

; set fuelbar color...:
    lda     fuel                ;               X = $ff!
.loopFuelCol:
    inx
    cmp     FuelValTbl,x
    bcs     .loopFuelCol
; ...and let low fuel blink:
    lda     FuelBlinkTbl,x
    beq     .skipFuelBlink
    and     frameCnt
    bne     .skipFuelBlink
    ldx     #4
.skipFuelBlink:
    ldy     FuelColTbl,x
    sty     fuelCol

; position BL at right end of fuelbar:
    lda     fuel
    and     #3                  ; 2             A = fuel!
    eor     #2                  ;               if (fuel & 3 == 2) x-pos--
    cmp     #1
    lda     fuel
    and     #$fc
    adc     #1
    jsr     CalcPosX
    sty     hMoveBL
    ora     hMoveBL
    sta     hMoveBL

; set tank missile size:
    ldy     dirMisl
    lda     MislSizeTbl,y       ; 4
    sta     NUSIZ1              ; 3 = 39

;    jsr     NextRandom
;    rts
;GameCalc END;


NextRandom SUBROUTINE
    lda     random              ; 3
    lsr                         ; 2
    bcc     .skipEor            ; 2
    eor     #RAND_EOR_8         ; 2
.skipEor:                       ;   = 11/12 (~11.5)
    sta     random              ; 3

  IF DISABLE
    lda    random               ; 3
    asl                         ; 2
    eor    random               ; 3
    asl                         ; 2
    eor    random               ; 3
    asl                         ; 2
    asl                         ; 2
    eor    random               ; 3
    asl                         ; 2
    rol    random               ; 5
  ENDIF
    rts                         ; 6
;NextRandom END


CheckJSR SUBROUTINE
    jmp     (jmpPtr)            ;~47


TankAI SUBROUTINE
.tankIdx    = tmpVar

;***** missile AI: *****
; check if missile active:
    lda     yPosMisl            ; 3
    bne     .skipMisl           ; 2=  5/6

; get tank orientation:
    lda     dirTankLst,x        ; 4
    lsr                         ; 2
    lsr                         ; 2
    tay                         ; 2
    lsr                         ; 2
    bcc     .vertDirTank        ; 2= 14/15

; check if helicopter is in vertical range:
    lda     yPosPlayer          ; 3
    sbc     yPosTankLst,x       ; 4             CF=1!
    adc     #4                  ; 2
    cmp     #9                  ; 2
    bcs     .skipMisl           ; 2= 13/14

; check if helicopter is in horzizontal range:
    lda     xPosTankLst,x       ; 4
    cmp     xPosPlayer          ; 3
    tya                         ; 2
    eor     #DIR_RIGHT/4        ; 2
    bpl     .contHorz           ; 2= 13/14
    DEBUG_BRK

; check if helicopter is in horzizontal range:
.vertDirTank:
    lda     xPosPlayer          ; 3
    sbc     xPosTankLst,x       ; 4             CF=0!
    adc     #4+1                ; 2
    cmp     #9                  ; 2
    bcs     .skipMisl           ; 2= 13/14

; check if helicopter is in vertical range:
    lda     yPosTankLst,x       ; 4
    cmp     yPosPlayer          ; 3
    tya                         ; 2 =  9
.contHorz:
    bne     .downLeftDirTank    ; 2
    bcc     .startMisl          ; 2
    clc                         ; 2
.downLeftDirTank:
    bcc     .skipMisl           ; 2=  5/8

.startMisl:
; fire new missile:
    sty     dirMisl             ; 3
    lda     xPosTankLst,x       ; 4
    adc     MislOfsX,y          ; 4
    sta     xPosMisl            ; 3
    lda     yPosTankLst,x       ; 4
    adc     MislOfsY,y          ; 4
    sta     yPosMisl            ; 3
    lda     MislHeight,y        ; 4
    sta     hMisl               ; 3
    stx     mislTank            ; 3             remember which tank fired missile
.skipMisl:
; total: 6..92


;***** find a new direction for the tank *****
.yGridTank  = tmpVar3

; check if tank is exactly at grid position:
;calc y-grid
    lda     yPosTankLst,x       ; 4
    ldy     #NUM_YBLOCKS*2      ; 2 =  6
.loopY:
    dey                         ; 2
    dey                         ; 2
    cmp     YGridTbl,y          ; 4
    bcc     .loopY              ; 2= 10/11
    bne     .skipTurn           ; 2
    sty     .yGridTank          ; 3 =  9

; calc x-grid
    lda     xPosTankLst,x       ; 4
    ldy     #NUM_XBLOCKS*2      ; 2 =  6
.loopX:
    dey                         ; 2
    dey                         ; 2
    cmp     XGridTbl,y          ; 4
    bcc     .loopX              ; 2= 10/11
    beq     .checkDirs
; total: 25..113

.skipTurn:
    rts

; get all valid new directions:
.validDirs  = tmpVar2
.dirIdx     = tmpVar4

.checkDirs:
    tya                         ; 2
    lsr                         ; 2
    tay

    ldx     #MAXDIR/4-1         ; 2 =  2
.loopCheckDir:
    stx     .dirIdx             ; 3
    lda     CheckJSRTbl,x       ; 4
    sta     jmpPtr              ; 3
    lda     CheckDirTbl,x       ; 4
    ldx     .tankIdx            ; 3
    cmp     dirTankLst,x        ; 4
    beq     .notDir             ; 2
    ldx     .yGridTank          ; 3
    jsr     CheckJSR            ;53             y = Y-GridTank
    bne     .notDir             ; 2
    sec                         ; 2
    NOP_B                       ; 0
.notDir:
    clc                         ; 2
    rol     .validDirs          ; 5
    ldx     .dirIdx             ; 3
    dex                         ; 2
    bpl     .loopCheckDir       ; 2= 38/39/97/98
; total: ~333 cylces

; convert distances to helicopter into index values (tricky!):
.diffX      = tmpVar3

    ldx     .tankIdx            ; 3
    ldy     #5                  ; 2
    lda     xPosPlayer          ; 3
    sec                         ; 2
    sbc     xPosTankLst,x       ; 4
    beq     .equX               ; 2
    bcs     .posX               ; 2
    ldy     #0                  ; 2
    sbc     #0                  ; 2             CF=0!
    eor     #$ff                ; 2
    NOP_W                       ; 0
.equX:
    ldy     #8                  ; 2
.posX:
    sta     .diffX              ; 3 = 22/29

    lda     yPosPlayer          ; 3
    sec                         ; 2
    sbc     yPosTankLst,x       ; 4
    beq     .equY               ; 2
    bcs     .posY               ; 2
    sbc     #0                  ; 2             CF=0!
    eor     #$ff                ; 2
    iny                         ; 2
    iny                         ; 2
.posY:
    iny                         ; 2
.equY:                          ;   = 12/16/23
    cmp     .diffX              ; 3             dY > dX?
    bcc     .skipIny            ; 2
    iny                         ; 2
.skipIny:                       ;   =  6/7
; total: 40..59

; select new direction based on valid and ordered directions:
.tmpDir     = tmpVar3

    ldx     #MAXDIR/4-1         ; 2
    lda     TurnDirTbl,y        ; 4 =  6
.loopAIDir:
    sta     .tmpDir             ; 3
    and     #%11                ; 2
    tay                         ; 2
    eor     random              ; 3
    and     RandomMaskTbl,x     ; 4             variable randomness (prefer optimal directions)
    beq     .skipDir            ; 2= 16/17
.skipRandom:
    lda     BitTbl,y            ; 3
    and     .validDirs          ; 3
    bne     .doTurn             ; 2=  8/9
.skipDir:
    lda     .tmpDir             ; 3
    lsr                         ; 2
    lsr                         ; 2
    dex                         ; 2
    bpl     .loopAIDir          ; 2= 11/12
; total: 31..148

; make sure that there is a valid direction:
    ldy   #-1                   ; 2 =  2
.loopTurn:
    iny                         ; 2
    lsr     .validDirs          ; 5
    bcc     .loopTurn           ; 2=  9/10
; total: 11..41

; continue with old or start turning into new direction:
.doTurn:
    ldx     .tankIdx            ; 3
    tya                         ; 2
    ora     dirTankLst,x        ; 4
    tay                         ; 2
    lda     NewDirTbl,y         ; 4
    sta     dirTankLst,x        ; 4
    rts                         ; 6 = 25
; max total: ~900 cycles
;TankAI END


CheckOtherTanks SUBROUTINE
; check if tank is close to other tanks
; x = tank-index
; y = direction/4
.tankIdx    = tmpVar
.xPosTank   = tmpVar2
.yPosTank   = tmpVar3

    lda     yPosTankLst,x       ; 4
    sta     .yPosTank           ; 3
    lda     xPosTankLst,x       ; 4
    sta     tmpVar2             ; 3
    ldx     #NUM_TANKS-1        ; 2 = 12
.loopTanks:
    cpx     .tankIdx            ; 3             don't check tank with itself
    beq     .nextTank           ; 2=  4/5

    lda     .xPosTank           ; 3
    clc                         ; 2
    adc     XCheck1Tbl,y        ; 4
    sbc     xPosTankLst,x       ; 4             CF=0!
    cmp     XCheck2Tbl,y        ; 4
    bcs     .nextTank           ; 2= 19/20

    lda     .yPosTank           ; 3
;    clc                         ; 2
    adc     YCheck1Tbl,y        ; 4             CF=0!
    sbc     yPosTankLst,x       ; 4             CF=0!
    cmp     YCheck2Tbl,y        ; 4
    bcc     .startUTurn         ; 2= 17/18
.nextTank:
    dex                         ; 2
    bpl     .loopTanks          ; 2=  4/5

    ldx     .tankIdx            ; 3
    rts                         ; 6 =  9

.startUTurn:
; revive other tank:
    lda     dirTankLst,x        ; 4
    and     #~NO_TANK           ; 2
    sta     dirTankLst,x        ; 4 = 10

; enable positive u-turn:
    ldx     .tankIdx            ; 2
    lda     dirTankLst,x        ; 4
    ora     #POS_TURN|U_TURN|1  ; 2
    sta     dirTankLst,x        ; 4
    rts                         ; 6 = 18
; total: 114..181
;CheckOtherTanks END


GetGridHigh SUBROUTINE
    lda     yGrid
    and     #~%1
    tax
    lda     xGrid
    lsr
    tay
    rts


DisableTankCheck SUBROUTINE
    cpy     mislTank
    beq     .rts
DisableTank:
; TODO: scoring
;disable tank:
    lda     dirTankLst,y
    ora     #NO_TANK
    sta     dirTankLst,y
.rts:
    rts

OverScan SUBROUTINE
    lda     #36+4
    sta     TIM64T

;********** collisions **********
; shot collisions:
; which tanks are displayed?
    jsr     GetFrameXY

    dey
    bit     CXM1P               ;               missile vs. tank?
    bmi     .disableTankCheck
    iny
    bvc     .loopShot
.disableTankCheck:
    jsr     DisableTankCheck

.loopShot:
    lda     CXP0FB,x            ;               shot vs. tank?
    asl
    bpl     .nextShot

    lda     dirShot             ;               check if tank has (nearly +/-1) same direction
    asl
    asl
    sta     tmpVar
    lda     dirTankLst,y        ;
    adc     #1
    and     #%1111
    sec
    sbc     tmpVar
    cmp     #3
    bcs     .nextShot
;    cmp     dirTankLst,y
;    bne     .nextShot
    jsr     DisableTank
    bne     .disableShot
    DEBUG_BRK

.nextShot:
    dey
    dex
    bpl     .loopShot

    lda     yPosShot
    cmp     ##7*H_VERT + 6*H_HORZ+1+2
    bcs     .disableShot
    bit     saveCXBLPF          ;               shot vs. PF?
    bpl     .skipDisableShot
.disableShot:
    lda     #0
    sta     yPosShot
.skipDisableShot:

; missile collisions:
    lda     yPosMisl
    cmp     ##7*H_VERT + 6*H_HORZ+1+2
    bcs     .disableMisl

    bit     CXM1FB              ;               missile vs. PF?
    bpl     .skipMisl           ;
.disableMisl:
    lda     #0
    sta     yPosMisl
.skipMisl:


;*******************************************
    inc     frameCnt

;*******************************************
    lda     killCnt
    bne     .skipJMP

    bit     CXM0P               ;               helicopter vs. tank?
    bvs     .killPlayer
    bmi     .killPlayer

    bit     CXPPMM              ;               helicopter vs. missile?
    bvc     .contGame
    lda     #0
    sta     yPosMisl
.killPlayer:
    lsr     lives
    lsr     lives
;    dec     lives
    lda     #64
    sta     killCnt
    bne     .skipJMP
    DEBUG_BRK

;*******************************************
.contGame:
    ldx     #NUM_TANKS-1
    lda     #NO_TANK
.loopCheckEnd:
    and     dirTankLst,x
    dex
    bpl     .loopCheckEnd
    tax
    beq     .doMoves

    ldx     maze
    inx
    cpx     #NUM_MAZES
    bcc     .okMazes
    ldx     ##0
.okMazes:
    stx     maze
    jsr     InitMaze
.skipJMP:
    jmp     .skipAllMoves

;********** move tanks **********
.tankIdx    = tmpVar
.loopCnt    = tmpVar5
; tankDir;
; 0..1: turning step
; 2..3: direction
; 6   : negative turning direction
; 7   : turning

.doMoves:
; increase tank delay only every second frame,
;  so that both tank groups will be moved:
    lda     frameCnt
    lsr
    bcc     .skipTankDelay
    lda     delayTank
    adc     speedTank
    sta     delayTank
.skipTankDelay:

;    lda     frameCnt
;    and     #$02
;    bne     .skipMoveTanks

    lda     delayTank
    cmp     speedTank
    bcc     .doMoveTanks
    eor     #$80
    cmp     speedTank
    bcs     .skipMoveTanks
.doMoveTanks:
    lda     #>CheckLeft
    sta     jmpPtr+1

; only check/move tanks visible in next frame:
    jsr     GetFrameXY          ;25
    stx     .loopCnt            ; 3
    tya                         ; 2
    tax                         ; 2 = 32
.loopMove:
    stx     .tankIdx            ; 3
    lda     dirTankLst,x        ; 4
    tay                         ; 2
    and     #(NO_TANK|U_TURN)|%11;2             tank disabled or turning?
    beq     .moveTank           ; 2= 13/14      no, move forward

;*** turn the tank: ***
    cmp     #NO_TANK            ;               tank disabled?
    bcs     .nextMove           ;                yes, skip turning

    lda     delayTank
    cmp     speedTank
    bcc     .nextMove

    tya
    and     #%11
    cmp     #%01

;    lda     frameCnt
;    and     #$04
;    bne     .nextMove

    tya
    bcs     .noUTurn
    and     #~U_TURN            ;               disable turn flag
    tay
.noUTurn:
    cpy     #NEG_TURN
    bcs     .turnNeg
    and     #%1111              ;               prevent bit-overflow
    eor     #%1111
    bne     .turnPos
    tya
    and     #POS_TURN|U_TURN
    tay
    NOP_B
.turnPos:
    iny
    NOP_B
.turnNeg:
    dey
    sty     dirTankLst,x
  IF ILLEGAL
    bne     .nextMove
    DEBUG_BRK
  ELSE
    jmp     .nextMove
  ENDIF

;---------------------------------------
.moveTank:                      ;14
    tya                         ; 2             tank finished turning?
    bmi     .skipAI             ; 2             yes, skip AI

    jsr     TankAI              ;900
    lda     dirTankLst,x        ; 4             tank started turning?
    bmi     .nextMove           ; 2             yes, don't move
.skipAI:
    and     #%1100              ; 2              disable turning flags
    sta     dirTankLst,x        ; 4
    lsr                         ; 2
    lsr                         ; 2
    tay                         ; 2
    txa                         ; 2
    adc     #ID_TANKLST         ; 2
    tax                         ; 2
    sec
    lda     speedTankHi
    jsr     MoveObjectDir       ;40
    ldx     .tankIdx            ; 3
    jsr     CheckOtherTanks     ;187
.nextMove:
    dex                         ; 2
    dec     .loopCnt
    bpl     .loopMove           ; 2
.skipMoveTanks:


; set helicopter graphic direction:
    ldx     dirPlayer
    lda     SWCHA
    bmi     .skipRightDir
    ldx     #%011
.skipRightDir:
    asl
    bmi     .skipLeftDir
    ldx     #%110
.skipLeftDir:
    stx     dirPlayer
    asl
    bmi     .skipDownDir
;    lda     dirPlayer
    txa
    and     #~%110
    ora     #%100
    bne     .setDirY

.skipDownDir:
    asl
    bmi     .skipUpDir
;    lda     dirPlayer
    txa
    and     #~%110
.setDirY:
    tax
.skipUpDir
    stx     dirPlayer


    bit     INPT4
    bmi     .skipFire
    lda     yPosShot
    bne     .skipFire
    lda     dirPlayer
    lsr
    sta     dirShot
    lda     xPosPlayer
    adc     #3
    sta     xPosShot
    lda     yPosPlayer
    sbc     #2
    sta     yPosShot
.skipFire:

;*****************************************************************
; move helicopter horizontal/vertical:
    lda     delayPlayer
    sec
    adc     speedPlayer
    sta     delayPlayer
    bcs     .doMoveHeli
;    lda     frameCnt
;    and     #$01
;    bne     .doMoveHeli
    jmp     .skipMoveHeli
;    jmp     .skipJoystick

.doMoveHeli:
    ldx     #yPosPlayer
    ldy     yGrid
    lda     YGridTbl,y
    bit     xyDir
    bmi     .moveVert
    bvc     .skipMoveHeli
    ldx     #xPosPlayer
    ldy     xGrid
    lda     XGridTbl,y
.moveVert:
    cmp     $00,x               ;       x/yPosPlayer
    bcc     .moveNeg
    inc     $00,x               ;       x/yPosPlayer
    NOP_W
.moveNeg:
    dec     $00,x               ;       x/yPosPlayer
    eor     $00,x               ;       x/yPosPlayer
    bne     .skipMoveHeli
    sta     xyDir               ;       A = 0!
.skipMoveHeli:

;---------------------------------------
; prefer change of helicopter direction:
    inc     xyDir
;    bit     xyDir
;    bvs     .checkVert
    cpx     #yPosPlayer
    bne     .checkVert

.checkHorz:
    bit     xyDir
    bmi     .endHorz
    lda     yGrid
    lsr
    bcs     .endHorz

    ldx     xGrid
    ldy     XGridTbl,x
    lda     SWCHA
    bmi     .skipRight
    cpy     xPosPlayer
    beq     .contRight
    bcs     .endHorz
.contRight:
    jsr     GetGridHigh
    jsr     CheckRight
    bne     .endHorz
    inc     xGrid
    bpl     .setHorz
    DEBUG_BRK
.skipRight:

    asl
    bmi     .skipLeft
    cpy     xPosPlayer
    bcc     .endHorz
    txa
    lsr
    bcs     .skipCheckLeft
    jsr     GetGridHigh
    jsr     CheckLeft
    bne     .endHorz
.skipCheckLeft:
    dec     xGrid
.setHorz:
    lda     #$40
    bne     .setMove

.skipLeft:
.endHorz:
    dec     xyDir

;---------------------------------------
.checkVert:
    bit     xyDir
    bvs     .endVert
    lda     xGrid
    lsr
    bcs     .endVert

    ldx     yGrid
    ldy     YGridTbl,x
    lda     SWCHA
    asl
    asl
    bmi     .skipDown
    cpy     yPosPlayer
    bcc     .endVert
    txa
    lsr
    bcs     .skipCheckDown
    jsr     GetGridHigh
    jsr     CheckDown
    bne     .endVert
.skipCheckDown:
    dec     yGrid
    bpl     .setVert
    DEBUG_BRK
.skipDown:

    asl
    bmi     .skipUp
    cpy     yPosPlayer
    beq     .contUp
    bcs     .endVert
.contUp:
    jsr     GetGridHigh
    jsr     CheckUp
    bne     .endVert
    inc     yGrid
.setVert:
    lda     #$80
.setMove:
    sta     xyDir
.skipUp:
.endVert:
    lda     xyDir
    lsr
    bcc     .skipJoystick
    jmp     .checkHorz

.skipJoystick:


.skipAllMoves:
    sta     CXCLR

  IF DEBUG
    jsr     GetGridHigh
    txa
    lsr
    sta     tmpVar
    tya
    asl
    asl
    asl
    asl
    ora     tmpVar
    sta     score
  ENDIF

    lda     frameCnt
    and     #$0f
    bne     .skipFuel
    dec     fuel
    bpl     .skipFuel
    lda     #MAX_FUEL
    sta     fuel
.skipFuel:
;    lda     #4
;    sta     fuel

.waitTim:
    lda     INTIM
    bne     .waitTim
    rts
;OverScan END


    align 256

CheckLeft SUBROUTINE            ;           ~42 cylces
    cpy     #1                  ; 2
    bcc     .notFree            ; 2
    dey                         ; 2
    jsr     GetPattern          ;24
    iny                         ; 2
    and     HorzMaskTbl-1,y     ; 4
.notFree:
    rts                         ; 6 = 11/42

CheckRight SUBROUTINE           ;           ~39 cylces
    cpy     #NUM_XBLOCKS-2      ; 2
    bcc     .checkFree          ; 2
    bne     .notFree            ; 2
.checkFree:
    jsr     GetPattern          ;24
    and     HorzMaskTbl,y       ; 4
.notFree:
    rts                         ; 6 = 13/39


CheckDown SUBROUTINE            ;           ~45 cycles
    cpx     #1                  ; 2
    bcc     NotFreeVert         ; 2
    dex                         ; 2
    jsr     GetPattern          ;24
    inx                         ; 2
    bne     ContCheckDown       ; 3 = 35

CheckUp SUBROUTINE              ;           ~42 cylces
    cpx     #NUM_YBLOCKS*2-3    ; 2
    bcs     NotFreeVert         ; 2
    inx                         ; 2
    jsr     GetPattern          ;24
    dex                         ; 2
ContCheckDown:
    and     VertMaskTbl,y       ; 4
NotFreeVert:
    rts                         ; 6 = 11/42


GetPattern SUBROUTINE           ;           ~17.5 cycles
    cpy     #6                  ; 2
    bcs     .leftRight          ; 2
GetPattern2:
    cpy     #2                  ; 2
    bcc     .leftRight          ; 2
    lda     PF2Lst,x            ; 4
    rts                         ; 6 = 18

.leftRight:
    lda     PF1Lst,x            ; 4
    rts                         ; 6 = 15/19
;GetPattern END


CalcPosX SUBROUTINE
; calculates the values and positions objects:
; Input:
; - a = x-position
    tay                         ; 2
    and    #$0F                 ; 2
    sta    tmpVar               ; 3
    tya                         ; 2
    lsr                         ; 2
    lsr                         ; 2
    lsr                         ; 2
    lsr                         ; 2
    tay                         ; 2
    clc                         ; 2
    adc    tmpVar               ; 3
    cmp    #$0F                 ; 2
    bcc    .skipIny             ; 2
    sbc    #$0F                 ; 2
    iny                         ; 2
.skipIny:
    eor    #$07                 ; 2
    asl                         ; 2
    asl                         ; 2
    asl                         ; 2
    asl                         ; 2
; Return:
; - y = coarse value for delay loop
; - a = fine value for HMxy
    rts
;CalcPosX END


SetPosX SUBROUTINE
    sta    WSYNC                ; 3
;---------------------------------------
    iny                         ; 7
    nop                         ; 2
    nop                         ; 2
    sta    HMP0,x               ; 4         @15
.waitPos:
    dey                         ; 2
    bpl    .waitPos             ; 2
.pageCheck:
    sta    RESP0,X              ; 4         @23
    rts                         ; 6
;SetPosX END

  IF >.waitPos != >.pageCheck
    ECHO ""
    ECHO "ERROR: SetPosX: >.waitPos != >.pageCheck !"
    ECHO ""
    ERR
  ENDIF


  IF PREDEFINED
DecodePF SUBROUTINE
.code   = tmpVar
.center = tmpVar2
.PFPtr  = tmpVar3

    lda     #<PFCodeTbl
    sta     .PFPtr
    lda     #>PFCodeTbl
    sta     .PFPtr+1

;set PF color:
    ldy     maze
    lda     PFMazeColTbl,y
    sta     COLUPF

    tya
    asl
    asl
    asl
    ora     #7
    tay
    lda     (.PFPtr),y
    sta     .center
    ldx     #NUM_BLOCKS-1
    bne     .enterLoop

.loopBlocks:
;horizontal:
    lda     #%10000000
    lsr     .code
    ror
.skipBitH0:
    sta     PF0Lst,x
    asl

    lda     PF1Lst+1,x
    bcc     .skipBitH1a
    ora     #%11100000
.skipBitH1a:
    lsr     .code
    bcc     .skipBitH1b
    ora     #%00111110
.skipBitH1b:
    lsr     .code
    bcc     .skipBitH1c
    ora     #%00000011
.skipBitH1c:
    sta     PF1Lst,x

    lda     PF2Lst+1,x
.loopNibble2h:
    bcc     .skipBitH2a
    ora     #%00000111
.skipBitH2a:
    lsr     .code
    bcc     .skipBitH2b
    ora     #%00111100
.skipBitH2b:
    lsr     .center
    bcc     .skipCenter
    ora     #%11000000
.skipCenter:
    sta     PF2Lst,x
    dex

;vertical:
.enterLoop:
    dey
    lda     (.PFPtr),y
    lsr
    sta     .code

    lda     #%01000000
    sta     PF0Lst,x

    lda     #0
    bcc     .skipBitV1a
    ora     #%00001000
.skipBitV1a:
    lsr     .code
    rol
    asl
    sta     PF1Lst,x
    ora     PF1Lst+1,x
    sta     PF1Lst+1,x

    lda     #0
    lsr     .code
    bcc     .skipBitV2a
    ora     #%00010000
.skipBitV2a:
    lsr     .code
    ror
    lsr
    sta     PF2Lst,x
    ora     PF2Lst+1,x
    sta     PF2Lst+1,x
    dex
    bpl     .loopBlocks
    rts
;ok

 echo "Decode ", . - DecodePF
  ENDIF


GenerateMaze SUBROUTINE

CheckFill:
.count      = tmpVar
.x1         = tmpVar3
.y1         = tmpVar4

    lda     #0                  ;           1st loop always finds adjacent points
.loopFill:
    sta     .count
    beq     .enterLoop          ;           jump directly to startpoint at 1st loop

    ldy     #3
.loopY:
    ldx     #NUM_BLOCKS-1       ;           12
.loopX:
    jsr     GetPattern2
    and     VertMaskTbl,y
    beq     .nextPoint
.enterLoop:
; check for left wall (*must* be before right check!):
    dey
    bmi     .skipLeft
    lda     #<CheckRight
    jsr     CheckDir
.skipLeft:
    iny

; check right:
    iny
    lda     #<CheckLeft
    jsr     CheckDir
    dey

; check down:
    dex
    bmi     .skipDown
    dex
    lda     #<CheckUp
    jsr     CheckDir
    inx
.skipDown:
    inx

; check up:
    cpx     #NUM_BLOCKS-1
    bcs     .skipUp
    inx
    inx
    lda     #<CheckDown
    jsr     CheckDir
    dex
    dex
.skipUp:

.nextPoint:
    dex
    dex
    bpl     .loopX
    dey
    bpl     .loopY

    tya
    cpy     .count
    bne     .loopFill

.endFill:
; clear all fill-markers:
    ldx     #NUM_BLOCKS-1
.loopClear:
    lda     PF1Lst,x
    and     #~%10001000
    sta     PF1Lst,x
    lda     PF2Lst,x
    and     #~%10010001
    sta     PF2Lst,x
    dex
    dex
    bpl     .loopClear

    ldx     .x1
    ldy     .y1
    rts
;CheckFill END


CheckDir:
    sta     jmpPtr

; check, if new point already set:
    lda     VertMaskTbl,y
    cpy     #2                  ; 2
    bcs     .readRight          ; 2
    and     PF1Lst,x            ; 4
    bcc     .readLeft           ; 2

.readRight:
    and     PF2Lst,x            ; 4
.readLeft:
    bne     .skipDir
    jsr     CheckJSR            ;~48
    bne     .skipDir

; check if we found the target point:
    cpx     .x1
    bne     .set
    cpy     .y1
    beq     .foundPoint

; now set the new point:
.set:
    sta     .count              ; 3
    lda     VertMaskTbl,y       ; 4
    cpy     #2                  ; 2
    bcc     .leftPF             ; 2
    ora     PF2Lst,x            ; 4
    sta     PF2Lst,x            ; 4
.skipDir;
    rts                         ; 6

.leftPF:
    ora     PF1Lst,x            ; 4
    sta     PF1Lst,x            ; 4
    rts                         ; 6

.foundPoint:
    pla
    pla
    clc
    bcc     .endFill
;CheckDir END


SetVert SUBROUTINE
; set vertical wall:
    lda     #%00000010
    cpy     #1
    bcc     .set0
    beq     .set1
    asl
    cpy     #2
    beq     .set2
    lda     #%01000000
.set2:
    eor     PF2Lst,x
    sta     PF2Lst,x
    rts

.set0:
    lda     #%00100000
.set1
    eor     PF1Lst,x
    sta     PF1Lst,x
    rts
;SetVert END


SetHorz SUBROUTINE
; set horizontal wall:
    cpy     #1
    beq     .set1
    bcs     .set234
.set0:
    lda     PF0Lst,x
    eor     #%10000000
    sta     PF0Lst,x
    lda     #%11100000
    NOP_W
.set1:
    lda     #%00111110
    eor     PF1Lst,x
    sta     PF1Lst,x
    rts

.set234:
    cpy     #3
    beq     .set3
    lda     #%11000000
    bcs     .set4
.set2:
    lda     #%00000011
    eor     PF1Lst,x
    sta     PF1Lst,x
    lda     #%00000111
    NOP_W
.set3:
    lda     #%00111100
.set4:
    eor     PF2Lst,x
    sta     PF2Lst,x
    rts
;SetHorz END


CreateMaze SUBROUTINE
; permanent local variables:
.wantedWalls    = tmpVar6
.totalWalls     = tmpVar7
.randomMask     = tmpVar8
; temporary local variables:
.forced         = tmpVar5
.x1             = tmpVar3
.y1             = tmpVar4
.tmpWalls       = tmpVar

    lda     #>CheckLeft
    sta     jmpPtr+1

    lda     #58
    sta     .totalWalls

    lda     #14
    bit     SWCHB
    bvs     .skipLow
    lda     #6
.skipLow:
    sta     .wantedWalls

    lda     #$3f
    sta     .randomMask

    ldx     #NUM_BLOCKS-1
.loopClear:
    lda     #%01000000
    sta     PF0Lst,x
    lda     #0
    sta     PF1Lst,x
    sta     PF2Lst,x
    dex
    bpl     .loopClear

    tay
.loopX:
; *** add horizontal walls: ***
    ldx     #11
.loopYHorz:
; check if point is solitaire:
; |
;-+?
; |
    inx
    dey
    bmi     .exitSIny
    jsr     CheckDown
    bne     .exitSIny
    iny
    jsr     CheckLeft
    bne     .exitSDex
    dex
    dex
    jsr     CheckLeft
    bne     .exitSInx
    inx                             ; forced walls can NOT partition maze,
    bpl     .undoHorz               ;  this also sets the wall

.exitSInx:
    inx
    NOP_W
.exitSIny
    iny
.exitSDex:
    dex

    jsr     CheckRandom
    bpl     .skipHorz

; check if horizontal wall doesn't produce dead end (max. 1 (or 0) borders allowed)
;+-
;|
;+?
;|
;+-
    inx
    jsr     CheckLeft
    beq     .checkLow
    cpy     #4
    beq     .exitDex
    jsr     CheckUp
    beq     .checkLow
.exitDex:
    dex
    NOP_B
.exitInx:
    inx
  IF ILLEGAL
    bpl     .skipHorz
    DEBUG_BRK
  ELSE
    jmp     .skipHorz
  ENDIF

.checkLow:
    dex
    dex
    jsr     CheckLeft
    beq     .setHorzInx
    cpy     #4
    beq     .exitInx
    jsr     CheckDown
    bne     .exitInx
.setHorzInx:
    inx
.setHorz:
    jsr     SetHorz

; check if new horizontal wall causes partitioned maze:
    cpy     #2                      ;           partition impossible in columns 0 and 1
    bcc     .skipHorzCheck

    dex
    stx     .x1
    sty     .y1
    inx
    inx
    jsr     CheckFill
    inx
    bcs     .undoHorz
.skipHorzCheck:
    dec     .wantedWalls
    bpl     .skipHorz
    DEBUG_BRK

.undoHorz:
    jsr     SetHorz
.skipHorz:
    dec     .totalWalls

; goto next row:
    dex
    dex
    bpl     .loopYHorz

; *** add vertical walls: ***
    cpy     #4                      ;           loop finished?
    bcs     .exitLoop               ;            yes, exit

    ldx     #NUM_BLOCKS-1
    bne     .checkVert              ; 3         first point is never solitaire

.loopYVert:
; check if point is solitaire:
; |
;-+
; ?
    jsr     CheckUp
    bne     .checkVert              ;            no, set normal random wall
    inx
    inx
    jsr     CheckRight
    php
    dex
    dex
    plp
    bne     .checkVert              ;            no, set normal random wall
    jsr     NextRandom16
    lda     randomHi
    bmi     .skipVert
    sec
    bpl     .setVert                ;            yes, set 50% random solitaire wall (forced)

.checkVert:
    jsr     CheckRandom
    bpl     .skipVert
;    clc
.setVert:
    ror     .forced

; check if vertical wall doesn't produce dead end (max. 1 walls allowed)
;+-+
;| ?
;+-+
    cpy     #1                      ;           no walls allowed here at bottom row
    beq     .specialVert
    cpy     #3                      ;           no walls allowed in center top and bottom column
    bne     .ok
    cpx     #12
    beq     .skipVert
.specialVert:
    txa
    beq     .skipVert
.ok:
    lda     #1
    sta     .tmpWalls

    jsr     CheckLeft
    beq     .notLeft
    dec     .tmpWalls
.notLeft:
    jsr     CheckUp
    beq     .notUp
    dec     .tmpWalls
.notUp:
    jsr     CheckDown
    beq     .notDown
    dec     .tmpWalls
.notDown:
    bit     .tmpWalls
    bmi     .skipVert

    jsr     SetVert

; check if new vertical wall causes partitioned maze:
    tya                             ;           partition impossible in column 0
    beq     .skipVertCheck

    stx     .x1
    sty     .y1
    iny
    jsr     CheckFill
    bcs     .undoVert
.skipVertCheck:
    bit     .forced
    bmi     .skipVert
    dec     .wantedWalls
    bpl     .skipVert
    DEBUG_BRK

.undoVert:
    jsr     SetVert
.skipVert:
    dec     .totalWalls

; goto next row:
    dex
    dex
    bpl     .loopYVert

; goto next column:
    iny
    jmp     .loopX                  ; 3

; fill all connection points:
.exitLoop:
    ldx     #NUM_BLOCKS-1-1
.loopPF12:
    lda     PF1Lst,x
    ora     #%00100010
    sta     PF1Lst,x
    lda     PF2Lst,x
    ora     PF2Lst+1,x
    dex
    bmi     .skip2
    ora     PF2Lst,x
.skip2:
    ora     #%00000100
    sta     PF2Lst+1,x
    dex
    bpl     .loopPF12

  IF DEBUG
    lda     .totalWalls             ;           should be 0
  ENDIF
    rts
;  0,404, 75 ( 6)
;  0,945, 20 (14) $23d
;  0,975, 58 (14) $227
;CreateMaze END


CheckRandom:
    lda     .totalWalls
    asl
    cmp     .randomMask
    bcs     .skipMask
    lsr     .randomMask         ;           adjust to optimize speed
.skipMask:

NextRandom16:
; implements a 16 bit LFSR which generates a new random number:
    lsr     randomHi            ; 5
    ror     randomLo            ; 5
    bcc     .skipEor            ; 2
    lda     randomLo            ; 3
    eor     #<RAND_EOR_16       ; 2
    sta     randomLo            ; 3
    lda     randomHi            ; 3
    eor     #>RAND_EOR_16       ; 2
    sta     randomHi            ; 3
.skipEor:                       ;
    lda     randomHi            ; 3 = 16/31 (~23.5)

    and     .randomMask
    cmp     .totalWalls
    bcc     .exitRandom
    sbc     .totalWalls
.exitRandom:
    cmp     .wantedWalls
    rts
;CheckRandom END

  echo "GenerateMaze ", . - GenerateMaze


;===============================================================================
; R O M - T A B L E S (Bank 0)
;===============================================================================

    org     $fd00, $ff

;    .byte %0000000
;    .byte %0000000
;    .byte %0000000
;    .byte %0000000
;    .byte %0000000
;    .byte %0000000
;    .byte %0000000

Game0:
    .byte %0111100
    .byte %1000100
    .byte %1001100
    .byte %1000000
    .byte %1000000
    .byte %1000100
    .byte %0111000

Game1:
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000

Game2:
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000
    .byte %0000000


  IF PREDEFINED
PFCodeTbl:
;maze #0:
    .byte   %0000
    .byte   %10100001
    .byte   %01000010
    .byte   %10000001
    .byte   %10000010
    .byte   %01000001
    .byte   %10100000
    .byte   %010010     ; center
;maze #1:
    .byte   %0100
    .byte   %00100000
    .byte   %00100110
    .byte   %00011000
    .byte   %10100010
    .byte   %00001101
    .byte   %01100000
    .byte   %110000     ; center
;maze #2:
    .byte   %0000
    .byte   %10100000
    .byte   %01010000
    .byte   %11000001
    .byte   %11000000
    .byte   %00001111
    .byte   %00000000
    .byte   %011100     ; center
;maze #3:
    .byte   %0000
    .byte   %01000001
    .byte   %11100001
    .byte   %10001010
    .byte   %01100010
    .byte   %10000001
    .byte   %00100100
    .byte   %100001     ; center
;maze #4:
    .byte   %0000
    .byte   %11100000
    .byte   %01001101
    .byte   %00101000
    .byte   %00101101
    .byte   %01000000
    .byte   %11100000
    .byte   %000000     ; center
;maze #5:
    .byte   %0000
    .byte   %00100100
    .byte   %11000101
    .byte   %00001010
    .byte   %00000101
    .byte   %11000100
    .byte   %00100000
    .byte   %100001     ; center
;maze #6:
    .byte   %0000
    .byte   %10100001
    .byte   %01001010
    .byte   %10000001
    .byte   %10001010
    .byte   %00100100
    .byte   %00100000
    .byte   %000001     ; center
;maze #7:
    .byte   %0000
    .byte   %10000011
    .byte   %00000100
    .byte   %00101000
    .byte   %00100100
    .byte   %00000011
    .byte   %10000000
    .byte   %010010     ; center
  ENDIF

; tables used for fuelbar:
FuelPF1Tbl:
    .byte   $00                                         ; wg PF0
    .byte   $00, $80, $c0, $e0, $f0, $f8, $fc, $fe
FuelPF2Tbl:
    .byte   $00, $01, $03, $07, $0f, $1f, $3f, $7f
FuelPFValTbl:
    .byte   0, 9,17,21
FuelPFOfsTbl
    .byte   0-1
    .byte   FuelPF2Tbl-FuelPF1Tbl-9-1
    .byte   FuelPF2Tbl-FuelPF1Tbl-17+4-1
    .byte   -21+1-1
FuelValTbl:
    .byte   13, 28, 58, MAX_FUEL+1
FuelBlinkTbl:
    .byte   $08, $18, $30, $00
FuelColTbl:
    .byte   $4a, $2a, $1a, FUEL_COL;, 0     shared with next table!
FuelBLSizeTbl:
    .byte   MSBL_SIZE1, MSBL_SIZE2, MSBL_SIZE4, MSBL_SIZE4

    align 256

  IF PREDEFINED
PFMazeColTbl:
    .byte   $6a, $a8, $c6, $08, $36, $96, $f8, $48
  ENDIF

VertMaskTbl:
    .byte   %10000000, %00001000                        ; $80, $10
    .byte   %00000001, %00010000, %10000000, %00010000  ; $01, $08, $80, $08
    .byte   $01, $10, $80
HorzMaskTbl:
    .byte   $20, $02
    .byte   $04, $40, $40, $04
    .byte   $02, $20

BlockSizeTbl:
    .byte   H_VERT*0 + H_HORZ*0
    .byte   H_VERT*1 + H_HORZ*0

    .byte   H_VERT*1 + H_HORZ*1
    .byte   H_VERT*2 + H_HORZ*1
    .byte   H_VERT*2 + H_HORZ*2
    .byte   H_VERT*3 + H_HORZ*2

    .byte   H_VERT*3 + H_HORZ*3
    .byte   H_VERT*4 + H_HORZ*3
    .byte   H_VERT*4 + H_HORZ*4
    .byte   H_VERT*5 + H_HORZ*4

    .byte   H_VERT*5 + H_HORZ*5
    .byte   H_VERT*6 + H_HORZ*5
    .byte   H_VERT*6 + H_HORZ*6


;****** SCORES ******
DH = 5

DigitPtrTbl:
    .byte   <Zero+DH-1-DigitTbl, <One+DH-1-DigitTbl, <Two  +DH-1-DigitTbl, <Three+DH-1-DigitTbl, <Four+DH-1-DigitTbl
    .byte   <Five+DH-1-DigitTbl, <Six+DH-1-DigitTbl, <Seven+DH-1-DigitTbl, <Eight+DH-1-DigitTbl, <Nine+DH-1-DigitTbl

DigitTbl:
Seven:
    .byte   %00100010
    .byte   %00100010
Four:
    .byte   %00100010
    .byte   %00100010
Zero:
    .byte   %11101110
    .byte   %10101010
    .byte   %10101010
    .byte   %10101010
Nine:
    .byte   %11101110
    .byte   %00100010
Eight:
    .byte   %11101110
    .byte   %10101010
Six:
    .byte   %11101110
    .byte   %10101010
    .byte   %11101110
    .byte   %10001000
Two:
    .byte   %11101110
    .byte   %10001000
Three:
    .byte   %11101110
    .byte   %00100010
Five:
    .byte   %11101110
    .byte   %00100010
    .byte   %11101110
    .byte   %10001000
One:
    .byte   %11101110
    .byte   %01000100
    .byte   %01000100
    .byte   %11001100
    .byte   %01000100

XGridTbl:
    .byte   X0,    X0+8,  X0+16,X0+24,X0+32, X0+40, X0+48, X0+55
    .byte   X0+62
    .byte   X0+69, X0+76, X0+84,X0+92,X0+100,X0+108,X0+116,X0+124
YGridTbl:
    .byte   Y0+NUM_BLOCKS*0, Y0+NUM_BLOCKS*1/2, Y0+NUM_BLOCKS*1, Y0+NUM_BLOCKS*3/2
    .byte   Y0+NUM_BLOCKS*2, Y0+NUM_BLOCKS*5/2, Y0+NUM_BLOCKS*3, Y0+NUM_BLOCKS*7/2
    .byte   Y0+NUM_BLOCKS*4, Y0+NUM_BLOCKS*9/2, Y0+NUM_BLOCKS*5, Y0+NUM_BLOCKS*11/2
    .byte   Y0+NUM_BLOCKS*6


; helicopter graphics:
PlayerSizeTbl:
    .byte   %100000
    .byte   %000000
    .byte   %100000
    .byte   %110000
    .byte   %100000
    .byte   %000000
    .byte   %100000
;    .byte   %100000
;    .byte   %000000
;    .byte   %110000
;    .byte   %100000
;    .byte   %010000
;    .byte   %100000
;    .byte   %100000

PlayerMoveLeftA:
    .byte   $10
    .byte   $e0
    .byte   $00
    .byte   $10
    .byte   $20
    .byte   $d0
    .byte   $f0
;    .byte   $20
;    .byte   $d0
;    .byte   $10
;    .byte   $20
;    .byte   $d0
;    .byte   $40
;    .byte   $b0
PlayerMoveLeftB:
    .byte   $10
    .byte   $e0
    .byte   $00
    .byte   $10
    .byte   $20
    .byte   $00
    .byte   $c0
;    .byte   $20
;    .byte   $d0
;    .byte   $10
;    .byte   $20
;    .byte   $10
;    .byte   $c0
;    .byte   $f0

PlayerMoveRightA:
    .byte   $20
    .byte   $f0
    .byte   $c0
    .byte   $30
    .byte   $10
    .byte   $d0
    .byte   $e0
PlayerMoveRightB:
    .byte   $20
    .byte   $f0
    .byte   $c0
    .byte   $30
    .byte   $10
    .byte   $00
    .byte   $b0

    align 256

; initial values for the tanks:
TankColTbl
    .byte   $c8, $3c, $8c, $4a
TankXPos:
    .byte   X_TANK0, X_TANK1, X_TANK2, X_TANK3
TankYPos:
    .byte   Y_TANK0, Y_TANK1, Y_TANK2, Y_TANK3
TankDir:
    .byte   DIR_TANK0, DIR_TANK1, DIR_TANK2, DIR_TANK3

Mul14Tbl:
    .byte   H_TANK*2*0, H_TANK*2*1, H_TANK*2*2, H_TANK*2*3
    .byte   H_TANK*2*4, H_TANK*2*5, H_TANK*2*6, H_TANK*2*7, H_TANK*2*8
    .byte   H_TANK*2*9


  IF >CheckUp != >CheckLeft
    ECHO ""
    ECHO "ERROR: >CheckUp != >CheckLeft !"
    ECHO ""
    ERR
  ENDIF

CheckJSRTbl:
    .byte   <CheckUp, <CheckRight, <CheckDown, <CheckLeft
CheckDirTbl:
    .byte   DIR_DOWN, DIR_LEFT, DIR_UP, DIR_RIGHT

UP_MASK    = %0001
RIGHT_MASK = %0010
DOWN_MASK  = %0100
LEFT_MASK  = %1000

BitTbl:
    .byte   UP_MASK, RIGHT_MASK, DOWN_MASK, LEFT_MASK       ; 1, 2, 4, 8

TurnDirTbl:
    .byte   DIR_LEFT /4|DIR_UP   |DIR_DOWN *4|DIR_RIGHT*16  ; 9

    .byte   DIR_LEFT /4|DIR_UP   |DIR_DOWN *4|DIR_RIGHT*16  ;10
    .byte   DIR_UP   /4|DIR_LEFT |DIR_RIGHT*4|DIR_DOWN *16  ;11
    .byte   DIR_LEFT /4|DIR_DOWN |DIR_UP   *4|DIR_RIGHT*16  ; 8
    .byte   DIR_DOWN /4|DIR_LEFT |DIR_RIGHT*4|DIR_UP   *16  ; 7

    .byte   DIR_RIGHT/4|DIR_DOWN |DIR_UP   *4|DIR_LEFT *16  ; 3

    .byte   DIR_RIGHT/4|DIR_UP   |DIR_DOWN *4|DIR_LEFT *16  ; 2
    .byte   DIR_UP   /4|DIR_RIGHT|DIR_LEFT *4|DIR_DOWN *16  ; 1
    .byte   DIR_RIGHT/4|DIR_DOWN |DIR_UP   *4|DIR_LEFT *16  ; 4
    .byte   DIR_DOWN /4|DIR_RIGHT|DIR_LEFT *4|DIR_UP   *16  ; 5

    .byte   DIR_UP   /4|DIR_RIGHT|DIR_LEFT *4|DIR_DOWN *16  ; 0
    .byte   0                                               ; -
    .byte   DIR_DOWN /4|DIR_LEFT |DIR_RIGHT*4|DIR_UP   *16  ; 6

RandomMaskTbl:
    .byte   $ff, $03, $07, $0f      ; 99.6%, 25%, 12.5%, 6.25%

NewDirTbl:
    .byte   DIR_UP        , POS_TURN|%0001, -1            , NEG_TURN|%1111
    .byte   NEG_TURN|%0011, DIR_RIGHT     , POS_TURN|%0101, -1
    .byte   -1            , NEG_TURN|%0111, DIR_DOWN      , POS_TURN|%1001
    .byte   POS_TURN|%1101, -1            , NEG_TURN|%1011, DIR_LEFT


XDirTbl:
    .byte   0; 1, 0,-1
YDirTbl:
    .byte   1, 0,-1, 0

; missile start offsets:
MISL_H  = 7
MISL_W  = 8

MislOfsX:
    .byte   4, 4, 3, 4-MISL_W
MislOfsY:
    .byte   MISL_H-4,-4,-3,-5
MislHeight
    .byte   MISL_H, 1, MISL_H, 1
MislSizeTbl:
    .byte   MSBL_SIZE1, MSBL_SIZE8, MSBL_SIZE1, MSBL_SIZE8

; boundaries for tank collisions:
XCheck1Tbl:
    .byte   H_TANK+1  , 10+1      , H_TANK+1  , 0+1
XCheck2Tbl:
    .byte   H_TANK*2+1, 10+1      , H_TANK*2+1, 10+1
YCheck1Tbl:
    .byte   10+1      , H_TANK+1  , 0+1       , H_TANK+1
YCheck2Tbl:
    .byte   10+1      , H_TANK*2+1, 10+1      , H_TANK*2+1

TankSpeedTbl:

HeliSpeedTbl:

MislSpeedTbl:

ShotSpeedTbl:

    org     . &$ff00 | $70

;****** PLAYERS ******
TankPatTbl:
Tank0a:
    .byte   %00000000
    .byte   %01000100
    .byte   %00101000   ; X X
    .byte   %10111010   ; XOX
    .byte   %00010000   ;  X
    .byte   %01000100
    .byte   %00010000   ;  X
Tank0b:
    .byte   %00000000
    .byte   %00000000
    .byte   %10101010
    .byte   %00111000
    .byte   %10010010
    .byte   %00000000
    .byte   %00111000
Tank23a:
    .byte   %00001000
    .byte   %00000000
    .byte   %10001010   ;   X
    .byte   %00111000   ; XXX
    .byte   %10011010   ;  XX
    .byte   %00000000
    .byte   %00101000   ;   X
Tank23b:
    .byte   %00000000
    .byte   %00000100
    .byte   %00001000
    .byte   %10111010
    .byte   %00011000
    .byte   %01000100
    .byte   %00011000
Tank45a:
    .byte   %00010000
    .byte   %00000100
    .byte   %00010000   ;  X
    .byte   %10111010   ; XXX
    .byte   %00011000   ;  XX
    .byte   %01000100   ;    X
    .byte   %00010000
Tank45b:
    .byte   %00001000
    .byte   %00000000
    .byte   %00010010
    .byte   %00111000
    .byte   %10011010
    .byte   %00000100
    .byte   %00101000
Tank78a:
    .byte   %00101000
    .byte   %00000000
    .byte   %00010010   ;  X
    .byte   %00011000   ;  XX
    .byte   %10111010   ; XXX X
    .byte   %00000000
    .byte   %00101000
Tank78b:
    .byte   %00010000
    .byte   %00000100
    .byte   %00010000
    .byte   %00011010
    .byte   %00111010
    .byte   %01000100
    .byte   %00010000
Tank90a
    .byte   %00010000
    .byte   %01000100
    .byte   %00110000   ; XX
    .byte   %00011010   ;  XX X
    .byte   %00110000   ; XX
    .byte   %01000100
    .byte   %00010000
Tank90b
    .byte   %00101000
    .byte   %00000000
    .byte   %00110010
    .byte   %00011010
    .byte   %00110010
    .byte   %00000000
    .byte   %00101000
Tank113a:
    .byte   %00101000
    .byte   %00000000
    .byte   %10111010   ; XXX X
    .byte   %00011000   ;  XX
    .byte   %00010010   ;  X
    .byte   %00000000
    .byte   %00101000
Tank113b:
    .byte   %00010000
    .byte   %01000100
    .byte   %00111010
    .byte   %00011010
    .byte   %00010000
    .byte   %00000100
    .byte   %00010000
Tank135a:
    .byte   %00010000
    .byte   %01000100   ;    X
    .byte   %00011000   ;  XX
    .byte   %10111010   ; XXX
    .byte   %00010000   ;  X
    .byte   %00000100
    .byte   %00010000
Tank135b:
    .byte   %00101000
    .byte   %00000100
    .byte   %10011010
    .byte   %00111000
    .byte   %00010010
    .byte   %00000000
    .byte   %00001000
Tank148a:
    .byte   %00101000   ;   X
    .byte   %00000000
    .byte   %10011010   ;  XX
    .byte   %00111000   ; XXX
    .byte   %10001010   ;   X
    .byte   %00000000
    .byte   %00001000
Tank148b:
    .byte   %00011000
    .byte   %01000100
    .byte   %00011000
    .byte   %10111010
    .byte   %00001000
    .byte   %00000100
    .byte   %00000000
Tank180a:
    .byte   %00010000   ;  X
    .byte   %01000100
    .byte   %00010000   ;  X
    .byte   %10111010   ; XOX
    .byte   %00101000   ; X X
    .byte   %01000100
    .byte   %00000000
Tank180b:
    .byte   %00111000
    .byte   %00000000
    .byte   %10010010
    .byte   %00111000
    .byte   %10101010
    .byte   %00000000
    .byte   %00000000

NoTankA:
    .byte   %00010000   ;   X
    .byte   %01000100   ; X   X
    .byte   %00000000   ;
    .byte   %10000010   ;X     X
    .byte   %00000000   ;
    .byte   %01000100   ; X   X
    .byte   %00010000   ;   X
NoTankB:
    .byte   %00101000   ;  X X
    .byte   %00000000   ;
    .byte   %10000010   ;X     X
    .byte   %00000000   ;
    .byte   %10000010   ;X     X
    .byte   %00000000   ;
    .byte   %00101000   ;  X X


    org $fffc
    .word   Start
    .word   Break